from unittest import TestCase

from pyramid.testing import DummyRequest
from pyramid.exceptions import ConfigurationError


class DummyLDAPConnector(object):

    def __init__(self, group_list):
        self.group_list = group_list

    # noinspection PyUnusedLocal
    def user_groups(self, userdn):
        return self.group_list


class Dummy(object):

    def __init__(self, *arg, **kw):
        pass


class DummyConfig(object):

    introspectable = Dummy

    def __init__(self):
        self.registry = Dummy()
        self.directives = []
        self.prop_reify = self.prop_name = self.prop = None

    # noinspection PyUnusedLocal
    def add_directive(self, name, directive):
        self.directives.append(name)

    def set_request_property(self, prop, name, reify=False):
        self.prop_reify = reify
        self.prop_name = name
        self.prop = prop

    # noinspection PyUnusedLocal
    @staticmethod
    def action(discriminator, action_task, introspectables=()):
        if action_task:
            action_task()


class DummyManager(object):

    def __init__(self, with_errors=None):
        self.with_errors = with_errors or []
        self.user = self.password = None
        self.status = self.exit_info = None

    def connection(self, user=None, password=None):
        self.user = user
        self.password = password
        if self.with_errors:
            e = self.with_errors.pop(0)
            if e is not None:
                raise e
        self.bind()
        return self

    def __enter__(self):
        pass

    def __exit__(self, *exit_info):
        self.exit_info = exit_info
        self.unbind()

    def bind(self):
        self.status = 'bound'
        return

    def unbind(self):
        self.status = 'unbound'


class DummySearch(object):

    def __init__(self, result, exc=None):
        self.result = result
        self.exc = exc
        self.conn = None
        self.kw = None

    def execute(self, conn, **kw):
        if self.exc is not None:
            raise self.exc
        self.conn = conn
        self.kw = kw
        return self.result


class DummyConnection(object):

    def __init__(self, result):
        self.result = result
        self.result_id = 0
        self.args = None
        self.kwargs = None

    def search(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        self.result_id += 1
        return self.result_id

    def get_response(self, result_id):
        if self.kwargs['search_scope'] == 'none':
            return None, result_id
        return self.result, result_id


class DummyLdap3Server(object):

    def __init__(self, host, port=None, use_ssl=False, tls=None):
        self.host = host
        self.port = port
        self.ssl = use_ssl
        self.tls = tls


class DummyLdap3Connection(object):

    def __init__(
            self, server, user=None, password=None,
            auto_bind=False, authentication=None,
            read_only=False, client_strategy=None, lazy=False,
            pool_name='pyramid_ldap3', pool_size=10, pool_lifetime=None):
        self.server = server
        self.user = user
        self.password = password
        self.auto_bind = auto_bind
        self.authentication = authentication
        self.read_only = read_only
        self.client_strategy = client_strategy
        self.lazy = lazy
        self.pool_name = pool_name
        self.pool_size = pool_size
        self.pool_lifetime = pool_lifetime


class DummyLdap3(object):

    STRATEGY_ASYNC_THREADED = 1
    STRATEGY_REUSABLE_THREADED = 2

    Server = DummyLdap3Server
    Connection = DummyLdap3Connection



import ldap3

from pyramid_ldap3 import groupfinder

from pyramid.config import Configurator
from pyramid.authentication import AuthTktAuthenticationPolicy
from pyramid.authorization import ACLAuthorizationPolicy
from pyramid.security import Allow, Authenticated


class RootFactory(object):

    __acl__ = [(Allow, Authenticated, 'view')]

    def __init__(self, request):
        pass


def main(global_config, **settings):
    config = Configurator(settings=settings, root_factory=RootFactory)
    config.include('pyramid_ldap3')
    config.include('pyramid_chameleon')
    config.set_authentication_policy(
        AuthTktAuthenticationPolicy('seekr1t', callback=groupfinder))
    config.set_authorization_policy(
        ACLAuthorizationPolicy())
    config.ldap_setup(
        'ldap://localhost',
        bind='CN=ldap user,CN=Users,DC=example,DC=com',
        passwd='ld@pu5er')
    config.ldap_set_login_query(
        'CN=Users,DC=example,DC=com',
        '(sAMAccountName=%(login)s)',
        scope=ldap3.SEARCH_SCOPE_WHOLE_SUBTREE,
        cache_period=0)
    config.ldap_set_groups_query(
        'CN=Users,DC=example,DC=com',
        # '(member:1.2.840.113556.1.4.1941:=%(userdn)s)',
        '(&(objectCategory=group)(member=%(userdn)s))',
        cache_period=60)
    config.add_route('sampleapp.root', '/')
    config.add_route('sampleapp.login', '/login')
    config.add_route('sampleapp.logout', '/logout')
    config.scan('.views')
    return config.make_wsgi_app()


from . import (
    TestCase, ConfigurationError,
    Dummy, DummyRequest, DummyManager, DummySearch)


class TestGetLdapConnector(TestCase):

    def _callFUT(self, request):
        from pyramid_ldap3 import get_ldap_connector
        return get_ldap_connector(request)

    def test_no_connector(self):
        request = DummyRequest()
        self.assertRaises(ConfigurationError, self._callFUT, request)
        
    def test_with_connector(self):
        request = DummyRequest()
        request.ldap_connector = True
        result = self._callFUT(request)
        self.assertEqual(result, True)


class TestConnector(TestCase):

    def _makeOne(self, registry, manager):
        from pyramid_ldap3 import Connector
        return Connector(registry, manager)

    def test_authenticate_no_ldap_login_query(self):
        manager = DummyManager()
        inst = self._makeOne(None, manager)
        self.assertRaises(ConfigurationError, inst.authenticate, None, None)
        self.assertTrue(manager.status is None)

    def test_authenticate_search_returns_no_result(self):
        manager = DummyManager()
        registry = Dummy()
        registry.ldap_login_query = DummySearch([])
        inst = self._makeOne(registry, manager)
        self.assertTrue(inst.authenticate(None, None) is None)
        self.assertEqual(manager.status, 'unbound')

    def test_authenticate_empty_password(self):
        manager = DummyManager()
        registry = Dummy()
        registry.ldap_login_query = DummySearch([('a', 'b')])
        inst = self._makeOne(registry, manager)
        self.assertTrue(inst.authenticate('foo', '') is None)
        self.assertTrue(manager.status is None)

    def test_authenticate_search_returns_one_result(self):
        manager = DummyManager()
        registry = Dummy()
        registry.ldap_login_query = DummySearch([('a', 'b')])
        inst = self._makeOne(registry, manager)
        self.assertEqual(inst.authenticate(None, None), ('a', 'b'))
        self.assertEqual(manager.status, 'unbound')

    def test_authenticate_search_returns_multiple_results(self):
        manager = DummyManager()
        registry = Dummy()
        registry.ldap_login_query = DummySearch([('a', 'b'), ('a', 'c')])
        inst = self._makeOne(registry, manager)
        self.assertTrue(inst.authenticate(None, None) is None)
        self.assertEqual(manager.status, 'unbound')

    def test_authenticate_search_bind_raises(self):
        from pyramid_ldap3 import ldap3
        manager = DummyManager([None, ldap3.LDAPException])
        registry = Dummy()
        registry.ldap_login_query = DummySearch([('a', 'b')])
        inst = self._makeOne(registry, manager)
        self.assertTrue(inst.authenticate(None, None) is None)
        self.assertEqual(manager.status, 'unbound')

    def test_user_groups_no_ldap_groups_query(self):
        manager = DummyManager()
        inst = self._makeOne(None, manager)
        self.assertRaises(ConfigurationError, inst.user_groups, None)
        self.assertTrue(manager.status is None)

    def test_user_groups_search_returns_result(self):
        manager = DummyManager()
        registry = Dummy()
        registry.ldap_groups_query = DummySearch([('a', 'b')])
        inst = self._makeOne(registry, manager)
        self.assertEqual(inst.user_groups(None), [('a', 'b')])
        self.assertEqual(manager.status, 'unbound')

    def test_user_groups_execute_raises(self):
        from pyramid_ldap3 import ldap3
        manager = DummyManager()
        registry = Dummy()
        registry.ldap_groups_query = DummySearch(
            [('a', 'b')], ldap3.LDAPException)
        inst = self._makeOne(registry, manager)
        self.assertTrue(inst.user_groups(None) is None)
        self.assertEqual(manager.status, 'unbound')

    def test_user_groups_search_escapes(self):
        manager = DummyManager()
        registry = Dummy()
        search = DummySearch([('a', 'b')])
        registry.ldap_groups_query = search
        inst = self._makeOne(registry, manager)
        self.assertEqual(inst.user_groups('abc123'), [('a', 'b')])
        self.assertEqual(search.kw['userdn'], 'abc123')
        self.assertEqual(inst.user_groups('(abc*123)'), [('a', 'b')])
        self.assertEqual(search.kw['userdn'], '\\28abc\\2A123\\29')
        self.assertEqual(inst.user_groups(b'ab\xc3\xa7123'), [('a', 'b')])
        self.assertEqual(search.kw['userdn'].encode('latin-1'), b'ab\xe7123')
        self.assertEqual(inst.user_groups(b'ab\xe7123'), [('a', 'b')])
        self.assertEqual(search.kw['userdn'], '\\61\\62\\e7\\31\\32\\33')
        self.assertEqual(manager.status, 'unbound')



from . import TestCase, DummyConnection


class TestLDAPQuery(TestCase):

    def _makeOne(self, base_dn, filter_tmpl, scope, attributes, cache_period):
        from pyramid_ldap3 import _LDAPQuery
        return _LDAPQuery(
            base_dn, filter_tmpl, scope, attributes, cache_period)

    def test_execute_no_result(self):
        inst = self._makeOne('DN=Org', '(cn=%(login)s)', 'none', 'attrs', 0)
        conn = DummyConnection([{'dn': 'a', 'attributes': {'b': 'c'}}])
        result = inst.execute(conn, login='foo')
        self.assertEqual(result, [])

    def test_query_cache_no_rollover(self):
        inst = self._makeOne(None, None, None, None, 1)
        inst.last_timeslice = 1 << 31
        inst.cache['foo'] = 'bar'
        self.assertEqual(inst.query_cache('foo'), 'bar')

    def test_query_cache_with_rollover(self):
        inst = self._makeOne(None, None, None, None, 1)
        inst.cache['foo'] = 'bar'
        self.assertTrue(inst.query_cache('foo') is None)
        self.assertEqual(inst.cache, {})
        self.assertNotEqual(inst.last_timeslice, 0)

    def test_execute_no_cache_period(self):
        inst = self._makeOne('DN=Org', '(cn=%(login)s)', 'scope', 'attrs', 0)
        conn = DummyConnection([{'dn': 'a', 'attributes': {'b': 'c'}}])
        result = inst.execute(conn, login='foo')
        self.assertEqual(inst.cache, {})
        self.assertEqual(result, [('a', {'b': 'c'})])
        self.assertEqual(conn.args, ('DN=Org', '(cn=foo)'))
        self.assertEqual(
            conn.kwargs, {'attributes': 'attrs', 'search_scope': 'scope'})

    def test_execute_with_cache_period_miss(self):
        inst = self._makeOne('DN=Org', '(cn=%(login)s)', 'scope', 'attrs', 1)
        conn = DummyConnection([{'dn': 'a', 'attributes': {'b': 'c'}}])
        result = inst.execute(conn, login='foo')
        self.assertEqual(result, [('a', {'b': 'c'})])
        self.assertEqual(conn.args, ('DN=Org', '(cn=foo)'))
        self.assertEqual(conn.kwargs, {
            'attributes': 'attrs', 'search_scope': 'scope'})

    def test_execute_with_cache_period_hit(self):
        inst = self._makeOne('DN=Org', '(cn=%(login)s)', 'scope', 'attrs', 1)
        inst.last_timeslice = 1 << 31
        inst.cache[('DN=Org', '(cn=foo)')] = ('d', {'e': 'f'})
        conn = DummyConnection([{'dn': 'a', 'attributes': {'b': 'c'}}])
        result = inst.execute(conn, login='foo')
        self.assertEqual(result, ('d', {'e': 'f'}))
        self.assertTrue(conn.args is None)
        self.assertTrue(conn.kwargs is None)



from . import TestCase, DummyRequest, DummyLDAPConnector


class TestGetGroups(TestCase):

    def _callFUT(self, dn, request):
        from pyramid_ldap3 import get_groups
        return get_groups(dn, request)

    def test_no_group_list(self):
        request = DummyRequest()
        request.ldap_connector = DummyLDAPConnector(None)
        result = self._callFUT('testdn', request)
        self.assertTrue(result is None)

    def test_with_group_list(self):
        request = DummyRequest()
        request.ldap_connector = DummyLDAPConnector([('a', 'b')])
        result = self._callFUT('testdn', request)
        self.assertEqual(result, [('a', 'b')])


class TestGroupfinder(TestCase):

    def _callFUT(self, dn, request):
        from pyramid_ldap3 import groupfinder
        return groupfinder(dn, request)

    def test_no_group_list(self):
        request = DummyRequest()
        request.ldap_connector = DummyLDAPConnector(None)
        result = self._callFUT('testdn', request)
        self.assertTrue(result is None)

    def test_with_group_list(self):
        request = DummyRequest()
        request.ldap_connector = DummyLDAPConnector([('groupdn', None)])
        result = self._callFUT('testdn', request)
        self.assertEqual(result, ['groupdn'])



from os import path
from setuptools import setup, find_packages

here = path.abspath(path.dirname(__file__))
README = open(path.join(here, 'README.rst')).read()
CHANGES = open(path.join(here, 'CHANGES.txt')).read()

requires = ['pyramid>=1.3', 'ldap3<2.0']

sampleapp_extras = ['waitress', 'pyramid_debugtoolbar']
testing_extras = ['nose', 'coverage']
docs_extras = ['Sphinx']

setup(
    name='pyramid_ldap3',
    version='0.2.5',
    description='pyramid_ldap3',
    long_description=README + '\n\n' + CHANGES,
    classifiers=[
        "Framework :: Pylons",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        'Topic :: Software Development :: Libraries :: Python Modules',
        "Topic :: System :: Systems Administration"
            " :: Authentication/Directory :: LDAP",
        "Intended Audience :: Developers",
        "Development Status :: 4 - Beta",
        "License :: Repoze Public License"],
    author='Chris McDonough, Christoph Zwerschke',
    author_email='pylons-discuss@groups.google.com',
    url='http://pylonsproject.org',
    license="BSD-derived (http://www.repoze.org/LICENSE.txt)",
    keywords='web pyramid pylons ldap auth authentication',
    packages=find_packages(exclude=['docs', 'tests', 'sampleapp']),
    zip_safe=False,
    install_requires=requires,
    tests_require=requires,
    extras_require=dict(
        sampleapp=sampleapp_extras,
        docs=docs_extras,
        testing=testing_extras),
    test_suite='tests',
    entry_points="""\
    [paste.app_factory]
    sampleapp = sampleapp:main
    """)
